#!/usr/bin/python3
from __future__ import absolute_import, print_function, unicode_literals
import os
import sys
import re

class DBconn:
    
    def __init__(self,dsn,dbtype='sqlite3',debug=False):
        """
        accepted options for dbtype: sqlite3 and pg (for postgres)
        """
        self.hdr  = "DBconn: "
        self.debug = debug
        self.dbtype = dbtype
        self.noclose = False
        self.dsn = dsn
        self.conn = None
        
        try:
            if dbtype == 'sqlite3':
                import sqlite3 as dbmodule
                self.conn = dbmodule.connect(dsn)
                self.conn.row_factory = dbmodule.Row
            elif dbtype == 'pg':
                import psycopg2 as dbmodule
                import psycopg2.extras
                self.conn = dbmodule.connect(dsn, cursor_factory=psycopg2.extras.DictCursor)
            else:
                print("Unrecognized DB type %s" % dbtype);
                self.noclose = True
                sys.exit()
        except:
            print("Cannot connect to database");
            sys.exit()
        
        self.dbmodule = dbmodule
        self.debug = debug
        # hide password; not used anywhere else
        self.dsn = re.sub( r"password\s*=\s*'(.*)'", "password='XXXXX'", self.dsn)
        
        # need to turn on foreign key explicitly for sqlite3!
        if dbtype == 'sqlite3':
            try:
                cur = self.conn.cursor()
                cur.execute("pragma foreign_keys = ON")
                cur.close()
            except self.dbmodule.Error as e:
                print("! Failed to turn on foreign key support!")
                
        self.create_tables()
        
        return
    
    def create_tables(self):
        """
        create tables in database (if they don't already exist)
        """
        cdir = os.path.dirname( os.path.realpath(__file__) )
        
        # table schemas -------------------------------------
        schema = os.path.join(cdir,"data","partsmaster.sql")
        if self.debug:
            print(self.hdr,"parts master schema is ",schema)
        
        self.populate(schema)
        
        schema = os.path.join(cdir,"data","approvedmfg.sql")
        if self.debug:
            print(self.hdr,"approved mfg list schema is ",schema)
        
        self.populate(schema)
        
        schema = os.path.join(cdir,"data","attachment.sql")
        if self.debug:
            print(self.hdr,"attachment schema is ",schema)
        
        self.populate(schema)

        schema = os.path.join(cdir,"data","bom.sql")
        if self.debug:
            print(self.hdr,"bill of materials schema is ",schema)
        
        self.populate(schema)
        
        return
        
    def __enter__(self):
        return self

    """
    # remove this for now; too much trouble reconciling this with flask
    def __del__(self):
        if self.debug:
            print(self.hdr,"closing database connection %s (auto2)" % self.dsn)
            self.close()
        return
    
    def __exit__(self,exc_type,exc_value,traceback):
        if self.debug:
            print(self.hdr,"closing database connection %s (auto1)" % self.dsn)
            self.close()
        return
    """
    
    def commit(self):
        self.conn.commit()
        return
        
    def close(self):
        if not self.noclose and self.conn != None:
            self.conn.close()
        return
    
    def populate(self,schema):
        
        with open(schema) as fh:
            mstr = fh.read()
        
        # NOTE: dict cursor use
        # dict_cur = self.conn( cursor_factory=psycopg2.extras.DictCursor )
        c = self.conn.cursor()
        c.execute(mstr)
        c.close()
        self.conn.commit()
        return
    
    def removeall(self):
        c = self.conn.cursor()
        err = False
        try:
            c.execute("delete from partsmaster")
            c.execute("delete from bom")
            c.execute("delete from attachment")
            c.execute("delete from approvedmfg")
        except self.db.dbmodule.Error as e:
            err = True
            errmsg = e
        
        if err:
            if self.dbtype == 'sqlite3':
                print("ERROR: ",errmsg)
                msg = errmsg
            else:
                print(errmsg.pgerror)
                print(errmsg.diag.message_detail)
                msg = errmsg.pgerror + "; " + errmsg.diag.message_detail
            
            c.close()
            self.conn.rollback()
            return msg
                
        self.conn.commit()
        return "ok"
    



#!/usr/bin/python3
from __future__ import absolute_import, print_function, unicode_literals
import sys
import os
import re

if __name__ == '__main__':
    cdir = os.path.dirname( os.path.realpath(__file__) )
    sys.path.insert(0, cdir+"/..")

import pypdx

def main():
    if len(sys.argv) < 3:
        print("USAGE: %s pdx-file.xml dsn [dump [remove_all_first]]" % sys.argv[0])
        print("     - pdx-file.xml: this is the PDX XML file")
        print("     - dsn: can be a SQLite3 file (the program will create one if it does not exist; use the extension .sqlite3")
        print("          : or it can be the dsn connection string for a PostgreSQL database")
        print("          : if dsn is 'pg', the default database 'pdx' on localhost (port 5432) and username 'pdxuser' will be used")
        print("     - dump: 1 or 0; to dump to a JSON file pdx-dump.json (optional)")
        print("     - remove_all: 1 or 0; remove all records from the tables first (optional);")
        print("     -           : WARNING: this will delete *all* existing parts, BOM, etc., records from the database")
        sys.exit()
    
    pdxfile = sys.argv[1]
    dsn     = sys.argv[2]
    
    if len(sys.argv) >= 4:
        dodump  = True if sys.argv[3]=='1' else False
    else:
        dodump = False

    if len(sys.argv) >= 5:
        removeall = True if sys.argv[4]=='1' else False
    else:
        removeall = False
    
    debug = True
    try:
        if dsn == 'pg':
            dsn = "dbname='pdx' user='pdxuser' host='localhost' port=5432"
            mypdx = pypdx.PDX(pdxfile,dsn, dbtype='pg',debug=debug)
        elif dsn[-8:] == '.sqlite3':
            mypdx = pypdx.PDX(pdxfile,dsn,debug=debug)
        elif re.match('dbname\s*=', dsn) != None:
            mypdx = pypdx.PDX(pdxfile,dsn, dbtype='pg',debug=debug)
        else:
            print("Unrecognized dsn %s" % dsn)
            sys.exit()
        
    except IOError as e:
        print("Connection Failed ",e)
        sys.exit(1)
    
    if dodump:
        mypdx.dump("pdx-dump.json")
    
    if removeall:
        if sys.version_info[0] < 3:
            doit = raw_input("MAIN: removing *all* records first! Are you sure? (type yes or no)")
        else:
            doit = input("MAIN: removing *all* records first! Are you sure? (type yes or no)")
        
        if doit == 'yes':
            mypdx.removeall()
        else:
            print("MAIN: not removing old records")
    
    status = mypdx.fillparts()
    if status == 'ok':
        print("MAIN: operation completed")
    else:
        print("MAIN: fillparts() failed")
        sys.exit(1)
    
    sys.exit(0)
    
# ==============================================
if __name__ == '__main__':
    main()




#!/usr/bin/python3
from __future__ import absolute_import, print_function, unicode_literals
import sys
import os
import xmltodict
import json

from . import dbconn

bomfields = [
             "revisionIdentifier",
             "isSerializationRequired",
             "globalBillOfMaterialTypeCode",
             "globalBillOfMaterialTypeCodeOther",
             "notes",
             "billOfMaterialItemIdentifier",
             "billOfMaterialItemUniqueIdentifier",
             "itemQuantity",
             "globalProductQuantityTypeCode",
             "globalProductQuantityTypeCodeOther",
             "description",
             "proprietarySequenceIdentifier",
            ]

attfields = [
             "referenceName",
             "universalResourceIdentifier",
             "fileIdentifier",
             "versionIdentifer",
             "fileSize",
             "checkSum",
             "isFileIn",
             "description",
             "globalMimeTypeQualifierCode",
             "attachmentModificationDate",
            ]

apmfields = [
             "manufacturerPartIdentifier",
             "manufacturerPartUniqueIdentifier",
             "manufacturerContactUniqueIdentifier",
             "globalManufacturerPartStatusCode",
             "globalManufacturerPartStatusCodeOther",
             "globalPreferredStatusCode",
             "description",
             "manufacturedBy",
            ]

# ===================================================================================
class PDX:
    def __init__(self,pdxfile, dsn, dbtype='sqlite3', debug=False):
        self.hdr = "PDX: "
        self.debug = debug
        self.dsn   = dsn
        
        # read into string then dump into object
        with open(pdxfile,'r') as myfile:
            xmldata = myfile.read()
        
        bomtree = self.bomtree = xmltodict.parse(xmldata)
        if not 'ProductDataeXchangePackage' in bomtree:
            print(self.hdr,"!!!! ERROR: XML file does not have ProductDataeXchangePackage")
            self.db = None
            return
        
        self.bomroot = bomtree['ProductDataeXchangePackage']
        
        # database connection
        self.dbtype = dbtype
        self.db = dbconn.DBconn(dsn, dbtype=dbtype, debug=debug)
        
        return
    
    def __del__(self):
        # self.db.close() # shouldn't need this if it doesn't crash!
        return
    
    # ---------------------------------------------------------------------------------------
    def dump(self,outputfile):
        if self.debug:
            print(self.hdr,"dump to file ",outputfile)
        
        with open(outputfile,"w") as fh:
            print(json.dumps( self.bomtree, indent=8, separators=(',',':') ), file=fh )
        
        return
    
    # ---------------------------------------------------------------------------------------
    def removeall(self):
        status = self.db.removeall()
        
        return status
    
    # ---------------------------------------------------------------------------------------
    def fillparts(self):
        items = self.bomroot['Items']['Item']
        
        fields = [
                  "itemIdentifier",
                  "itemUniqueIdentifier",
                  "globalLifeCyclePhaseCode",
                  "globalLifeCyclePhaseCodeOther",
                  "globalProductTypeCode",
                  "itemClassification",
                  "revisionIdentifier",
                  "versionIdentifer",
                  "proprietaryProductFamily",
                  "category",
                  "globalProductUnitOfMeasureCode",
                  "makeBuy",
                  "makeBuyOther",
                  "minimumShippableRevision",
                  "revisionReleasedDate",
                  "revisionIncorporatedDate",
                  "isSerializationRequired",
                  "isCertificationRequired",
                  "ownerName",
                  "ownerContactUniqueIdentifier",
                  "isTopLevel",
                  "description",
                  ]
        
        cur = self.db.conn.cursor()
        if self.dbtype == 'sqlite3':
            sqltemp = "insert into partsmaster ("+(",".join(fields))+") values ("+(",".join( ["?" for x in fields] ) )+")"
        else:
            # postgres/psycopg2
            sqltemp = "insert into partsmaster ("+(",".join(fields))+") values ("+(",".join( ['%s' for x in fields] ) )+")"
            
        # print(self.hdr,"Template: ",sqltemp)
        
        self.bomsqllist = [] # list of tuples to enter in bom table; initialize empty
        self.attsqllist = [] # list of tuples to enter in attachment table; initialize empty
        self.apmsqllist = [] # list of tuples to enter in approvedmfg table; initialize empty
        
        count = 0
        for item in items:
            if self.debug:
                print(self.hdr,"-",item['@description'])
            vlist = []
            for field in fields:
                if ('@'+field in item):
                    val = item['@'+field]
                    # timestamps: use the date-time string in all cases
                    
                    # boolean
                    if field[:2] == 'is':
                        if self.dbtype == 'sqlite3':
                            val = 1 if val == 'Yes' else 0
                        else:
                            val = 'True' if val == 'Yes' else 'False'
                else:
                    # val = 'NULL'
                    val = None
                vlist.append( val )
            
            # print(self.hdr," values: ",vlist)
            err = False
            try:
                cur.execute(sqltemp, tuple(vlist))
            except self.db.dbmodule.Error as e:
                err = True
                errmsg = e
            
            if err:
                if self.dbtype == 'sqlite3':
                    print("ERROR: ",errmsg)
                    msg = errmsg
                else:
                    print(errmsg.pgerror)
                    print(errmsg.diag.message_detail)
                    msg = errmsg.pgerror + "; " + errmsg.diag.message_detail
                
                cur.close()
                self.db.conn.rollback()
                return msg
            
            # .......... look for BOM ..............
            if 'BillOfMaterial' in item.keys():
                self.process_bom( item )
                
            # .......... look for Attachment ..............
            if 'Attachments' in item.keys():
                self.process_attachment( item )

            # .......... look for Approved MFG ..............
            if 'ApprovedManufacturerList' in item.keys():
                self.process_appmfg( item )

            count += 1

        cur.close()
        self.db.commit()
        
        # ------ process BOM list -----------------------------------
        cur = self.db.conn.cursor()
        if self.dbtype == 'sqlite3':
            sqltemp = "insert into bom (itemUniqueIdentifier,"+(",".join(bomfields))+") values (?,"+(",".join( ["?" for x in bomfields] ) )+")"
        else:
            # postgres/psycopg2
            # sqltemp = "insert into partsmaster ("+(",".join([('"%s"' % x) for x in fields]))+") values ("+(",".join( ['%s' for x in fields] ) )+")"
            sqltemp = "insert into bom (itemUniqueIdentifier,"+(",".join(bomfields))+") values (%s,"+(",".join( ['%s' for x in bomfields] ) )+")"
        
        for record in self.bomsqllist:
            err = False
            try:
                cur.execute(sqltemp, record)
            except self.db.dbmodule.Error as e:
                err = True
                errmsg = e
            
            if err:
                if self.dbtype == 'sqlite3':
                    print("ERROR: ",errmsg)
                    msg = errmsg
                else:
                    print(errmsg.pgerror)
                    print(errmsg.diag.message_detail)
                    msg = errmsg.pgerror + "; " + errmsg.diag.message_detail
                
                cur.close()
                self.db.conn.rollback()
                return msg
        
        cur.close()
        self.db.commit()
        
        # ------ process Attachment list -----------------------------------
        cur = self.db.conn.cursor()
        if self.dbtype == 'sqlite3':
            sqltemp = "insert into attachment (itemUniqueIdentifier,"+(",".join(attfields))+") values (?,"+(",".join( ["?" for x in attfields] ) )+")"
        else:
            # postgres/psycopg2
            sqltemp = "insert into attachment (itemUniqueIdentifier,"+(",".join(attfields))+") values (%s,"+(",".join( ['%s' for x in attfields] ) )+")"
        
        for record in self.attsqllist:
            err = False
            try:
                cur.execute(sqltemp, record)
            except self.db.dbmodule.Error as e:
                err = True
                errmsg = e
            
            if err:
                if self.dbtype == 'sqlite3':
                    print("ERROR: ",errmsg)
                    msg = errmsg
                else:
                    print(errmsg.pgerror)
                    print(errmsg.diag.message_detail)
                    msg = errmsg.pgerror + "; " + errmsg.diag.message_detail
                
                cur.close()
                self.db.conn.rollback()
                return msg
        
        
        cur.close()
        self.db.commit()
        
        # ------ process Approved MFG list -----------------------------------
        cur = self.db.conn.cursor()
        if self.dbtype == 'sqlite3':
            sqltemp = "insert into approvedmfg (itemUniqueIdentifier,"+(",".join(apmfields))+") values (?,"+(",".join( ["?" for x in apmfields] ) )+")"
        else:
            # postgres/psycopg2
            sqltemp = "insert into approvedmfg (itemUniqueIdentifier,"+(",".join(apmfields))+") values (%s,"+(",".join( ['%s' for x in apmfields] ) )+")"
        
        for record in self.apmsqllist:
            err = False
            try:
                cur.execute(sqltemp, record)
            except self.db.dbmodule.Error as e:
                err = True
                errmsg = e
            
            if err:
                if self.dbtype == 'sqlite3':
                    print("ERROR: ",errmsg)
                    msg = errmsg
                else:
                    print(errmsg.pgerror)
                    print(errmsg.diag.message_detail)
                    msg = errmsg.pgerror + "; " + errmsg.diag.message_detail
                
                cur.close()
                self.db.conn.rollback()
                return msg
        
        cur.close()
        self.db.commit()

        # --------- final tally --------------------------------------
        if self.debug:
            print(self.hdr,"Total %d items" % count)
            print(self.hdr,"Total %d BOM links" % len(self.bomsqllist) )
            print(self.hdr,"Total %d Attachment items" % len(self.attsqllist) )
            print(self.hdr,"Total %d Approved MFG items" % len(self.apmsqllist) )
        
        return 'ok'
    
    # =========================================================================================
    def process_bom(self, item):
        # print("item: ",item['@description'])
        parentid = item['@itemUniqueIdentifier']
        
        # sanity check
        if not ('BillOfMaterialItem' in item['BillOfMaterial'].keys()):
            print("!!!",item['@description']," missing BillOfMaterialItem")
            return
        
        xref = item['BillOfMaterial']['BillOfMaterialItem']
        if isinstance(xref,list):
            self.process_bom_list(parentid, xref)
        elif isinstance(xref,dict):
            self.process_bom_item(parentid, xref)
        else:
            print("!!!Don't know how to handle BOM ",type(xref)," type")
            return
        
        return
    
    # ---------------------------------------------------------------------------------------
    def process_bom_list(self, parentid, blist):
        for bitem in blist:
            self.process_bom_item(parentid,bitem)
        
        return
    # ---------------------------------------------------------------------------------------
    def process_bom_item(self, parentid, bitem):
        itemrecord = [parentid]
        
        for field in bomfields:
            field = '@'+field
            if field in bitem:
                val = bitem[field]
                if field[:2] == 'is':
                    if self.dbtype == 'sqlite3':
                        val = 1 if val == 'Yes' else 0
                    else:
                        val = 'True' if val == 'Yes' else 'False'
                elif field == '@itemQuantity':
                    val = float(val)
            else:
                val = None
            
            itemrecord.append( val )
        
        self.bomsqllist.append( tuple(itemrecord) )
        # print("...",tuple(itemrecord))
        return

    # =========================================================================================
    def process_attachment(self, item):
        # print("item: ",item['@description'])
        parentid = item['@itemUniqueIdentifier']
        
        # sanity check
        if not ('Attachment' in item['Attachments'].keys()):
            print("!!!",item['@description']," missing Attachment")
            return
        
        xref = item['Attachments']['Attachment']
        if isinstance(xref,list):
            self.process_att_list(parentid, xref)
        elif isinstance(xref,dict):
            self.process_att_item(parentid, xref)
        else:
            print("!!!Don't know how to handle Attachment ",type(xref)," type")
            return
        
        return
    
    # ---------------------------------------------------------------------------------------
    def process_att_list(self, parentid, blist):
        for bitem in blist:
            self.process_att_item(parentid,bitem)
        
        return
    # ---------------------------------------------------------------------------------------
    def process_att_item(self, parentid, bitem):
        itemrecord = [parentid]
        
        for field in attfields:
            field = '@'+field
            if field in bitem:
                val = bitem[field]
                if field[:2] == 'is':
                    if self.dbtype == 'sqlite3':
                        val = 1 if val == 'Yes' else 0
                    else:
                        val = 'True' if val == 'Yes' else 'False'
                elif field == '@fileSize':
                    val = int(val)
            else:
                val = None
            
            itemrecord.append( val )
        
        self.attsqllist.append( tuple(itemrecord) )
        # print("...",tuple(itemrecord))
        return
    
    # =========================================================================================
    def process_appmfg(self, item):
        # print("item: ",item['@description'])
        parentid = item['@itemUniqueIdentifier']
        
        # sanity check
        if not ('ApprovedManufacturerListItem' in item['ApprovedManufacturerList'].keys()):
            print("!!!",item['@description']," missing ApprovedManufacturerList")
            return
        
        xref = item['ApprovedManufacturerList']['ApprovedManufacturerListItem']
        if isinstance(xref,list):
            self.process_apm_list(parentid, xref)
        elif isinstance(xref,dict):
            self.process_apm_item(parentid, xref)
        else:
            print("!!!Don't know how to handle ApprovedManufacturerListItem ",type(xref)," type")
            return
        
        return
    
    # ---------------------------------------------------------------------------------------
    def process_apm_list(self, parentid, blist):
        for bitem in blist:
            self.process_apm_item(parentid,bitem)
        
        return
    # ---------------------------------------------------------------------------------------
    def process_apm_item(self, parentid, bitem):
        itemrecord = [parentid]
        
        for field in apmfields:
            field = '@'+field
            if field in bitem:
                val = bitem[field]
            else:
                val = None
            
            itemrecord.append( val )
        
        self.apmsqllist.append( tuple(itemrecord) )
        # print("...",tuple(itemrecord))
        return




"""
This program takes a PDX XML file and extracts the information into
a relational database (choice of SQLite3 or PostgreSQL).
"""
from .parser import PDX
from .main import main
__version__ = "0.0.1b3"



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
    
# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
      name='pypdx',
      version='0.0.1b3',
      
      description='Extract information from a PDX XML file into a SQLite or PostgreSQL database',
      long_description=long_description,
      url='https://github.com/sid5432/pypdx',
      author='Sidney Li',
      author_email='sidneyli5432@gmail.com',
      classifiers=[  # Optional
                   # How mature is this project? Common values are
                   #   3 - Alpha
                   #   4 - Beta
                   #   5 - Production/Stable
                   'Development Status :: 3 - Alpha',
                   
                   # Indicate who your project is intended for
                   'Intended Audience :: Manufacturing',
                   'Topic :: Utilities',
                   
                   # Pick your license as you wish
                   'License :: OSI Approved :: Apache Software License',
                   
                   # Specify the Python versions you support here. In particular, ensure
                   # that you indicate whether you support Python 2, Python 3 or both.
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                  ],
      
      keywords='PDX package data exchange IPC IPC-2571 IPC-2570',
      
      # You can just specify package directories manually here if your project is
      # simple. Or you can use find_packages().
      #
      # Alternatively, if you just want to distribute a single Python file, use
      # the `py_modules` argument instead as follows, which will expect a file
      # called `my_module.py` to exist:
      #
      #   py_modules=["my_module"],
      #
      # packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
      packages=find_packages(exclude=['archive']),
      
      install_requires=['psycopg2','xmltodict'],
      
      extras_require={
                      # 'dev': ['check-manifest'],
                      'test': ['pytest'],
                     },
      
      # this doesn't seem to be doing anything
      package_data={
                    'pypdx': ['data/*.sql','data/*.xml'],
                   },
      include_package_data=True,
      
      # data_files=[('data', ['pypdx/data/partsmaster.sql',
      #                      'pypdx/data/bom.sql',
      #                      'pypdx/data/approvedmfg.sql',
      #                      'pypdx/data/attachment.sql',
      #                      'pypdx/data/init_pg_db.sql',
      #                      'pypdx/data/pdx-example.xml',
      #                     ])],
      
      # To provide executable scripts, use entry points in preference to the
      # "scripts" keyword. Entry points provide cross-platform support and allow
      # `pip` to create the appropriate form of executable for the target
      # platform.
      #
      # For example, the following would provide a command called `sample` which
      # executes the function `main` from this package when invoked:
      entry_points={  # Optional
                    'console_scripts': [
                                        'pypdx=pypdx:main',
                                       ],
                   },
     )
